from .current_liquidity_profile import CurrentLiquidityProfile
from .pool_profile import PoolProfile
from .position_profile import MatchedPositionProfile
from .target_liquidity_profile import TargetLiquidityProfile


from typing import Union, List
import json
from decimal import Decimal
from datetime import datetime
from ..utils.converter import convert_to_probability
from ..utils.bucketing import DEFAULT_AMERICAN_SCHEMA, DEFAULT_DECIMAL_SCHEMA, DEFAULT_PROBABILITY_SCHEMA
import time

class Market():

    def __init__(
        self,
        state,
        id: int,
        event_id: int,
        currency: str,
        pool_id: int,
        expiry_time: datetime,
        status: str,
        is_main_market: bool, 
        handicap: Union[Decimal, None],
        type_id: int,
        type_name: str,
        num_winners: int,
        num_selections: int,
        min_post_quantity: Decimal,
        supports_in_play: bool,
        in_play_delay_seconds: Decimal,
        is_cross_matching: bool,
        matched_volume: Decimal,
        available_volume: Decimal,
        terms: str,
        description: str,
        category_id: int,
        subcategory_id: int,
        payoff_signature_length: int,
        price_schema: str
    ):
        self.state = state
        self.id = id
        self.event_id = event_id 
        self.currency = currency 
        self.pool_id = pool_id
        self.expiry_time = expiry_time 
        self.status = status 
        self.is_main_market= is_main_market
        self.handicap = handicap 
        self.type_id = type_id 
        self.type_name= type_name
        self.num_winners = num_winners 
        self.num_selections= num_selections
        self.min_post_quantity = min_post_quantity 
        self.supports_in_play = supports_in_play 
        self.in_play_delay_seconds = in_play_delay_seconds 
        self.is_cross_matching = is_cross_matching 
        self.matched_volume= matched_volume
        self.available_volume= available_volume
        self.terms = terms
        self.description = description
        self.category_id = category_id
        self.subcategory_id = subcategory_id
        self.payoff_signature_length = payoff_signature_length
        self._price_schema = price_schema
        self._update_timestamp = time.time()

    @classmethod
    def load_from_json(
        cls,
        state,
        data: Union[str, dict]
    ):
        if type(data) == str:
            data = json.loads(data)
        return cls(
            state = state,
            id = data.get('id'),
            event_id = data.get('event'),
            currency = data.get('currency'),
            pool_id = data.get('pool'),
            expiry_time =  datetime.strptime(data.get('expiry_time_utc'), "%Y-%m-%dT%H:%M:%SZ") if data.get('expiry_time_utc') != None else None,
            status = data.get('status'),
            is_main_market = data.get('is_main_market'),
            handicap = data.get('handicap'),
            type_id = data.get('type').get('id') if data.get('type') else None,
            type_name = data.get('type').get('name') if data.get('type') else None,
            num_winners = data.get('num_winners'),
            num_selections = data.get('num_selections'),
            min_post_quantity = Decimal(data.get('min_post_quantity')) if data.get('min_post_quantity') != None else None,
            supports_in_play = data.get('supports_in_play'),
            in_play_delay_seconds = data.get('in_play_delay_seconds'),
            is_cross_matching = data.get('is_cross_matching'),
            matched_volume = Decimal(data.get('matched_volume')) if data.get('matched_volume') != None else None,
            available_volume = Decimal(data.get('available_volume')) if data.get('available_volume') != None else None,
            terms = data.get('terms'),
            description = data.get('description'),
            category_id = data.get('category'),
            subcategory_id = data.get('subcategory'),
            price_schema = data.get('price_schema'),
            payoff_signature_length = data.get('payoff_signature_length')
        )
        
    @classmethod
    def load_from_client(
        cls,
        state,
        id: int
    ):
        market_data = state.client.get_markets(
            market_ids = [id]
        )[0]
        return cls.load_from_json(
            state = state,
            data = market_data
        )

    @property
    def pool(self):
        return self.state.pools.get(
            self.pool_id
        )

    @property
    def selections(self):
        return list(filter(
            lambda s: s.market_id == self.id,
            self.state.selections.values()
        ))
   
    @property
    def orders(self):
        return list(filter(
            lambda o: o.market_id == self.id,
            self.state.orders.values()
        ))

    @property
    def check_selections_complete(self):
        return (len(self.selections) == self.num_selections)

    @property
    def price_schema(self):
        if self._price_schema.upper() == 'DEFAULT_PROBABILITY_SCHEMA':
            return DEFAULT_PROBABILITY_SCHEMA
        elif self._price_schema.upper() == 'DEFAULT_DECIMAL_SCHEMA':
            return DEFAULT_DECIMAL_SCHEMA
        elif self._price_schema.upper() == 'DEFAULT_AMERICAN_SCHEMA':
            return DEFAULT_AMERICAN_SCHEMA
        else:
            raise Exception

    def __repr__(self):
        return f"<Market: {self.id}>"

    def __str__(self):
        return f"Market: {self.id}"

    
    


"""
Client Module
-------------
This module contains BaseClient and TradingClient classes.
The BaseClient handles authentication and basic request wrappers.
The TradingClient extends the base client to include functionality
required to source listing information and to interact with 
the platform (managing orders) for an authenticated user account.


"""

from .base_client import BaseClient
from .trading_client import TradingClient



from .enums import *
from decimal import Decimal
import json
from datetime import datetime
from typing import List
from ..utils.encoder import OrderEncoder


class ReduceOrderArgs():

    def __init__(
        self,
        order_id: int,
        new_quantity: Decimal,
        quantity_format: QuantityFormat,
    ):
        self.order = order_id
        self.new_quantity = new_quantity
        self.quantity_format = quantity_format

    @classmethod
    def load_from_json(
        self,
        data: str,
    ):
        if type(data) == str:
            data = json.loads(data)
        self.id = data.get('id')
        self.new_quantity = Decimal(data.get('new_quantity'))
        self.quantity_format = QuantityFormat(data.get('quantity_format'))
    
    def serialize_to_json(
        self,
    ):
        return json.dumps(
            self.__dict__,
            cls=OrderEncoder
        )
    
    def serialize_to_dict(
        self,
    ):
        d = json.loads(
            self.serialize_to_json()
        )
        keys = [k for k in d.keys()]
        for k in keys:
            if d.get(k) == None:
                d.pop(k)
        return d




from ..instructions import NewOrderArgs, OrderSide
from typing import Union, List
import json
from .profiles import CurrentLiquidityProfile
from .profiles import MatchedPositionProfile
from decimal import Decimal
import numpy as np
from ..utils.converter import convert_to_probability
from ..utils.bucketing import DEFAULT_AMERICAN_SCHEMA, DEFAULT_DECIMAL_SCHEMA, DEFAULT_PROBABILITY_SCHEMA
import time

class Pool():

    def __init__(
        self,
        state,
        id: int,
        scenario_space: str,
        currency: str,
        payoff_signature_length: int,
        price_schema: str,
        status: str,
        min_residual_quantity: Decimal,
        min_post_quantity: Decimal,
        supports_in_play: bool,
        in_play_delay_seconds: Decimal,
        event_id: int,
    ):
        self.state = state
        self.id = id
        self.scenario_space = scenario_space
        self.currency = currency
        self.payoff_signature_length = payoff_signature_length
        self._price_schema = price_schema
        self.status = status
        self.min_residual_quantity = min_residual_quantity
        self.min_post_quantity = min_post_quantity
        self.supports_in_play = supports_in_play
        self.in_play_delay_seconds = in_play_delay_seconds
        self.event_id = event_id
        self._update_timestamp = time.time()

        # Profiles
        self.refresh_profiles(perform_regardless=True)

        # Checks on whether profile refreshes are required
        self._orders_updated = False
        self._positions_updated = False
        self._selections_updated = False
        self._markets_updated = False
        


    @property
    def is_frozen(
        self,
    ):
        if self._frozen == False:
            return False
        else:
            if (
                self._new_orders_to_thaw == [] and
                self._cancel_orders_to_thaw == [] and
                self._reduce_orders_to_thaw == []
            ):
                self._frozen = False
            if (
                self._freeze_time and 
                (time.time() - self._freeze_time) > 15
            ):
                print(f"Freeze time elapsed for pool {self.id}")
                self._frozen = False

        return self._frozen


    def refresh_profiles(
        self,
        perform_regardless = False
    ):

        if perform_regardless:
            self.refresh_matched_position_profile()
            self.refresh_current_liquidity_profile()
        else:
            if self._orders_updated:
                self.refresh_current_liquidity_profile()
                self._orders_updated = False
            if self._positions_updated:
                self.refresh_matched_position_profile()
                self._positions_updated = False
            # if self._selections_updated:
            #     self.refresh_arb_free_prices()
            #     self._selections_updated = False
            if self._markets_updated:
                self._markets_updated = False
                pass

    def refresh_current_liquidity_profile(
        self,
    ):
        self.current_liquidity_profile = CurrentLiquidityProfile.generate_from_orders(
            pool = self
        )
        # self.current_liquidity_profile.print()
        return self.current_liquidity_profile


    def refresh_matched_position_profile(
        self,
    ):
        self.matched_position_profile = MatchedPositionProfile.generate_from_positions(
            pool = self
        )
        # self.matched_position_profile.print()
        return self.matched_position_profile


    @classmethod
    def load_from_json(
        cls,
        state,
        data: Union[str, dict]
    ):
        if type(data) == str:
            data = json.loads(data)
        return cls(
            state = state,
            id = data.get('id'),
            scenario_space = data.get('scenario_space'),
            currency = data.get('currency'),
            payoff_signature_length = data.get('payoff_signature_length'),
            price_schema = data.get('price_schema'),
            status = data.get('status'),
            min_residual_quantity = Decimal(data.get('min_residual_quantity')),
            min_post_quantity = Decimal(data.get('min_post_quantity')),
            supports_in_play = data.get('supports_in_play'),
            in_play_delay_seconds = Decimal(data.get('in_play_delay_seconds')),
            event_id = data.get('event')
        )
    

    def update_from_json(
        self,
        data: Union[str, dict]
    ):
        if type(data) == str:
            data = json.loads(data)
        self.scenario_space = data.get('scenario_space')
        self.currency = data.get('currency')
        self.payoff_signature_length = data.get('payoff_signature_length')
        self.price_schema = data.get('price_schema')
        self.status = data.get('status')
        self.min_residual_quantity = Decimal(data.get('min_residual_quantity'))
        self.min_post_quantity = Decimal(data.get('min_post_quantity'))
        self.supports_in_play = data.get('supports_in_play')
        self.in_play_delay_seconds = Decimal(data.get('in_play_delay_seconds'))
        self.event_id = data.get('event')
        

    @classmethod
    def load_from_client(
        cls,
        state,
        id: int
    ):
        pool_data = state.client.get_pools(
            pool_ids = [id]
        )[0]
        return cls.load_from_json(
            state = state,
            data = pool_data
        )

    @property
    def markets(self):
        return list(filter(
            lambda m: m.pool_id == self.id,
            self.state.markets.values()
        ))

    @property
    def selections(self):
        selections = []
        for m in self.markets:
            selections += m.selections
        return selections

    @property
    def orders(self):
        return list(filter(
            lambda o: o.pool_id == self.id,
            self.state.orders.values()
        ))

    @property
    def positions(self):
        positions = []
        for s in self.selections:
            position = self.state.positions.get(s.id)
            if position:
                positions += [position]
        return positions

    @property
    def price_schema(self):
        if self._price_schema.upper() == 'DEFAULT_PROBABILITY_SCHEMA':
            return DEFAULT_PROBABILITY_SCHEMA
        elif self._price_schema.upper() == 'DEFAULT_DECIMAL_SCHEMA':
            return DEFAULT_DECIMAL_SCHEMA
        elif self._price_schema.upper() == 'DEFAULT_AMERICAN_SCHEMA':
            return DEFAULT_AMERICAN_SCHEMA
        else:
            raise Exception
    
    @property
    def matched_exposure_vector(
        self,
    ):
        return self.matched_position_profile.calculate_exposure_vector()

    @property
    def unmatched_exposure_vector(
        self,
    ):
        return self.current_liquidity_profile.calculate_exposure_vector()
        
    @property
    def matched_and_unmatched_exposure_vector(
        self,
    ):
        return np.add(
            self.matched_exposure_vector,
            self.unmatched_exposure_vector
        )
    
    @property
    def max_liability(
        self
    ):
        return  Decimal(-1 * np.min(self.matched_and_unmatched_exposure_vector))


    def marginal_collateral_required_for_new_orders(
        self,
        new_orders: List[NewOrderArgs]
    ):
        new_matched_and_unmatched_exposure_vector = np.copy(self.matched_and_unmatched_exposure_vector)
        for order in new_orders:
            selection = self.state.selections.get(order.selection_id)
            if selection != None:
                probability_price = convert_to_probability(order.price, order.price_format)
                if order.side == OrderSide.BUY:
                    order_exposure_vector = selection.payoff_signature.generate_net_payoff_vector(
                        price = probability_price
                    ) * order.payout_quantity_remaining
                else:
                    order_exposure_vector = selection.payoff_signature.invert().generate_net_payoff_vector(
                        price= (Decimal(1)-probability_price)
                    ) * order.payout_quantity_remaining
            
            order_exposure_vector[
                order_exposure_vector > 0
            ] = Decimal(0)

            new_matched_and_unmatched_exposure_vector = np.add(
                new_matched_and_unmatched_exposure_vector,
                order_exposure_vector
            )

        new_max_liability = -1 * np.min(new_matched_and_unmatched_exposure_vector)

        marginal_capital_required = new_max_liability - self.max_liability 

        return marginal_capital_required




    def __repr__(self):
        return f"<Pool: {self.id}>"

    def __str__(self):
        return f"Pool: {self.id}"


from decimal import Decimal
from ..order import Order, OrderSide
from typing import List
import numpy as np


class PoolProfile():

    def __init__(
        self,
        pool
    ):
        self.data = {}
        self.pool = pool

    def __setitem__(
        self,
        index, # (market_id, selection_id, buy/sell, price)
        value
    ):
        market_id, selection_id, side, price = index
        # if side not in [OrderSide.BUY, OrderSide.SELL]:
        #     raise Exception
        if market_id not in self.data.keys():
            self.data[market_id] = {}
        if selection_id not in self.data[market_id].keys():
            self.data[market_id][selection_id] = {
                OrderSide.BUY: {},
                OrderSide.SELL: {}
            }
        self.data[market_id][selection_id][side][price] = value
   
    def __getitem__(
        self,
        index, # (market_id, selection_id, buy/sell, price)
    ):
        # print(index)
        try:
            market_id, selection_id, side, price = index
            # print(market_id, selection_id, side, price)
            return self.data[market_id][selection_id][side][price]
        except:
            return None
   
    def add(
        self,
        market_id: int,
        selection_id: int,
        side: str,
        price: Decimal,
        quantity
    ):
        current_quantity = self[(market_id, selection_id, side, price)]
        if current_quantity is None:
            self[(market_id, selection_id, side, price)] = quantity
        else:
            self[(market_id, selection_id, side, price)] = current_quantity + quantity

    def subtract(
        self,
        market_id: int,
        selection_id: int,
        side: str,
        price: Decimal,
        quantity
    ):
        current_quantity = self[(market_id, selection_id, side, price)]
        if current_quantity is None:
            self[(market_id, selection_id, side, price)] = -quantity
        else:
            self[(market_id, selection_id, side, price)] = current_quantity - quantity

    def remove(
        self,
        market_id: int,
        selection_id: int,
        side: str,
        price: Decimal,
        quantity
    ):
        current_quantity = self[(market_id, selection_id, side, price)]
        if current_quantity is not None:
            self[(market_id, selection_id, side, price)] = current_quantity - quantity


    @property
    def market_ids(
        self,
    ):
        return list(self.data.keys())

    @property
    def selection_ids(
        self,
    ):
        selection_ids = []
        for market_id in self.market_ids:
            selection_ids += list(self.data[market_id].keys())
        return selection_ids


    @classmethod
    def generate_delta_from_two_profiles(
        cls,
        target_profile,
        starting_profile
    ):
        if target_profile.pool != starting_profile.pool:
            raise Exception

        delta_profile = cls(
            pool=target_profile.pool
        )
        actions = []

        for market_id in target_profile.data.keys():
            for selection_id in target_profile.data[market_id].keys():
                for side in target_profile.data[market_id][selection_id].keys():
                    for price in target_profile.data[market_id][selection_id][side].keys():
                        delta_profile.add(
                            market_id = market_id, 
                            selection_id = selection_id,
                            side = side,
                            price = round(price, 5),
                            quantity = target_profile.data[market_id][selection_id][side][price]
                        )

        for market_id in starting_profile.data.keys():
            for selection_id in starting_profile.data[market_id].keys():
                for side in starting_profile.data[market_id][selection_id].keys():
                    for price in starting_profile.data[market_id][selection_id][side].keys():
                        delta_profile.subtract(
                            market_id = market_id, 
                            selection_id = selection_id,
                            side = side,
                            price = round(price, 5),
                            quantity = starting_profile.data[market_id][selection_id][side][price]
                        )

        for market_id in delta_profile.data.keys():
            for selection_id in delta_profile.data[market_id].keys():
                for side in delta_profile.data[market_id][selection_id].keys():
                    for price in delta_profile.data[market_id][selection_id][side].keys():
                        
                        delta_quantity = delta_profile[market_id, selection_id, side, price] 
                        starting_quantity = starting_profile[market_id, selection_id, side, price] 
                        starting_quantity = 0 if starting_quantity == None else starting_quantity
                        action = None

                        if abs(delta_quantity) < Decimal(0.01): #i.e. no material change
                            pass # no action required
                        elif delta_quantity > 0: # Increase only
                            action = {
                                'type': 'new',
                                'market_id': market_id,
                                'selection_id': selection_id,
                                'side': side,
                                'price': price,
                                'quantity': delta_quantity
                            }
                        else: # i.e. delta_quantity < 0:
                            if abs(starting_quantity - delta_quantity) < Decimal(0.01):
                                action = {
                                    'type': 'remove',
                                    'market_id': market_id,
                                    'selection_id': selection_id,
                                    'side': side,
                                    'price': price,
                                    'quantity': delta_quantity
                                }
                            else: # Just a reduction
                                action = {
                                    'type': 'reduce',
                                    'market_id': market_id,
                                    'selection_id': selection_id,
                                    'side': side,
                                    'price': price,
                                    'quantity': delta_quantity
                                }

                        if action:
                            actions += [action]

        return delta_profile, actions

    def calculate_exposure_vector(
        self,
    ):
        unmatched_exposure_vector = np.asarray([Decimal(0) for i in range(self.pool.payoff_signature_length)])
        for market_id in self.data.keys():
            for selection_id in self.data[market_id].keys():
                for side in self.data[market_id][selection_id].keys():
                    for price, quantity in sorted(self.data[market_id][selection_id][side].items()):
                        selection = self.pool.state.selections.get(selection_id)
                        if side == OrderSide.BUY:
                            selection_side_exposure = selection.payoff_signature.generate_negative_only_net_payoff_vector(
                                price = price 
                            ) * quantity
                        else:
                            selection_side_exposure = selection.payoff_signature.invert().generate_negative_only_net_payoff_vector(
                                price = Decimal(1)-price
                            ) * quantity
                        
                        unmatched_exposure_vector = np.add(
                            unmatched_exposure_vector,
                            selection_side_exposure
                        )

        return unmatched_exposure_vector


    def print(
        self
    ):
        print('='*40)
        print('POOL PROFILE')
        print(f'Pool: {self.pool.id}')
        print('-'*40)
        for market_id in self.data.keys():
            print(f'MARKET: {market_id}')
            for selection_id in self.data[market_id].keys():
                print(f' - SELECTION: {selection_id}')
                for side in [OrderSide.SELL, OrderSide.BUY]: #self.data[market_id][selection_id].keys():
                    print(f'   - {side.value.upper()}')
                    ticks = sorted(self.data[market_id][selection_id][side].items())
                    if side == OrderSide.BUY:
                        ticks = ticks[::-1]
                    for price, quantity in ticks:
                        print(f'      {price:.3f}: {quantity:8.2f}')
            print('-'*40)
        print('='*40)








from .exceptions import *
import requests
import json

class BaseClient():

    """
    The Base Client handles user authentication and connection to the Goated API.
    A connection can be authenticated using either:
    - An account email and password via BaseClient.create_with_login()
    - An authentication token via BaseClient.create_with_token()
    """

    def __init__(
        self,
        url: str,
        token: str = None,
        api_key: str = None,
        api_secret: str = None
    ):
        self._url = url if url[-1] == '/' else url + '/'
        self._token = token
        self._api_key = api_key
        self._api_secret = api_secret
        if self._token or (self._api_key and self._api_secret):
            self._generate_header()
        else:
            self._header = None

    @classmethod
    def create_with_login(
        cls,
        url: str,
        email: str,
        password: str
    ):
        '''
        Creates a BaseClient connection when provided with a valid Goated API URL, email and password.
        '''
        client = cls(
            url = url
        )
        client.login(
            email = email,
            password = password
        )
        return client

    @classmethod
    def create_with_token(
        cls,
        url: str,
        token: str,
    ):
        '''
        Creates a BaseClient connection when provided with a valid Goated API URL and valid authentication token (returned from the auth/login endpoint).
        '''
        return cls(
            url = url,
            token = token
        )

    @classmethod
    def create_with_api_key(
        cls,
        url: str,
        api_key: str,
        api_secret: str
    ):
        '''
        Creates a BaseClient connection when provided with a valid Goated API URL and valid API Key and API Secret.
        '''
        return cls(
            url = url,
            api_key = api_key,
            api_secret = api_secret
        )

    def _generate_header(
        self,
    ):
        if self._token:
            self._header = {
                'Authorization': 'Token ' + self._token
            }
        elif self._api_key and self._api_secret:
            self._header = {
                'GOATED-API-KEY': self._api_key,
                'GOATED-API-SECRET': self._api_secret
            }
        else:
            raise Exception()

    def _get_objects(
        self,
        endpoint: str,
        filters: dict,
        get_all_paginated: bool = True
    ):
        '''
        Gets a list of objects from the API, including iterating through pagination until all items are obtained.
        '''
        objects = []
        response = self._send_get(
            endpoint=endpoint,
            params = filters
        )
        objects += response.get('results')
        if response.get('next') is not None and get_all_paginated:
            while response.get('next'):
                response = self._send_get(
                    endpoint=response.get('next'),
                )
                objects += response.get('results')
        return objects

    def _get_object(
        self,
        endpoint: str,
    ):
        '''
        Gets a single object from the API.
        '''
        response = self._send_get(
            endpoint=endpoint,
        )

        return response

    def login(
        self,
        email: str,
        password: str,
    ):
        '''
        Attempts to login via the Goated Authentication API.
        '''
        response = self._send_post(
            endpoint = "auth/login/",
            data = {
                'email': email,
                'password': password
            }
        )   
        self._token = response.get('key')
        self._generate_header()


    def _send_get(
        self,
        endpoint: str,
        data: dict = {},
        params: dict = {}
    ):  
        '''
        Assembles and executes an authenticated GET request.
        '''
        if str(self._url) in endpoint:
            endpoint = endpoint.replace(str(self._url), '')
        response = requests.get(
            self._url + endpoint,
            json = data,
            headers= self._header,
            params = params
        )
        # print(response)
        if response.status_code not in [200, 201, 202]:
            raise GoatedApiException(response)
        json_data = json.loads(response.text)
        if json_data.get('success') == True:
            return json_data.get('data')
        else:
            raise GoatedApiException(json_data)

    def _send_post(
        self,
        endpoint: str,
        data: str = {},
        params: dict = {}
    ):
        '''
        Assembles and executes an authenticated POST request.
        '''
        if str(self._url) in endpoint:
            endpoint = endpoint.replace(str(self._url), '')
        response = requests.post(
            self._url + endpoint,
            json = data,
            headers= self._header,
            params = params
        )
        if response.status_code not in [200, 201, 202]:
            raise GoatedApiException(response)
        json_data = json.loads(response.text)
        if json_data.get('success') == True:
            return json_data.get('data')
        else:
            raise GoatedApiException(json_data)
        

    def _send_patch(
        self,
        endpoint: str,
        data: str = {},
        params: dict = {}
    ):
        '''
        Assembles and executes an authenticated PATCH request.
        '''
        if str(self._url) in endpoint:
            endpoint = endpoint.replace(str(self._url), '')
        response = requests.post(
            self._url + endpoint,
            json = data,
            headers= self._header,
            params = params
        )
        if response.status_code not in [200, 201, 202]:
            raise GoatedApiException(response)
        json_data = json.loads(response.text)
        if json_data.get('success') == True:
            return json_data.get('data')
        else:
            raise GoatedApiException(json_data)
        


from enum import Enum

class OrderSide(Enum):
    BUY = "buy"
    SELL = "sell"

class OrderType(Enum):
    LIMIT = "limit"
    MARKET = "market"
    IOC = "ioc"
    FOK = "fok"
    POST_ONLY = "post_only"

class PriceFormat(Enum):
    PROBABILITY = "probability"
    DECIMAL = "decimal"
    AMERICAN = "american"

class QuantityFormat(Enum):
    PAYOUT = "payout"
    STAKE = "stake"

class PersistenceType(Enum):
    INCLUDE_IN_AUCTION = "include_in_auction"
    LAPSE = "lapse"
    PERSIST = "persist"

class OrderStatus(Enum):
    OPEN = "open"
    PENDING = "pending"
    CLOSED = "closed"
    FAILED = "failed"



from typing import Union, List
import json
from ..utils.payoff_signature import PayoffSignature
from decimal import Decimal
import time

class Selection():

    def __init__(
        self,
        state,
        id: int,
        market_id: int,
        name: str,
        competitor: Union[dict, None],
        resolution_status: Union[str, None],
        payoff_signature: str,
        display_order: int,
        status: str,
        price_probability: str
    ):
        self.state = state
        self.id = id
        self.market_id = market_id
        self.name = name
        self.competitor = competitor
        self.resolution_status = resolution_status
        self.payoff_signature = PayoffSignature.from_string(payoff_signature)
        self.display_order = display_order
        self.status = status
        self.price_probability = Decimal(price_probability) if price_probability else Decimal(1)
        self._update_timestamp = time.time()

    @classmethod
    def load_from_json(
        cls,
        state,
        data: Union[str, dict]
    ):
        if type(data) == str:
            data = json.loads(data)
        return cls(
            state = state,
            id = data.get('id'),
            market_id = data.get('market'),
            name = data.get('name'),
            competitor = data.get('competitor'),
            resolution_status = data.get('resolution_status'),
            payoff_signature = data.get('payoff_signature'),
            display_order = data.get('display_order'),
            status =data.get('status'),
            price_probability = data.get('price_probability'),
        )
        
    @classmethod
    def load_from_client(
        cls,
        state,
        id: int
    ):
        selection_data = state.client.get_selections(
            selection_ids = [id]
        )[0]
        return cls.load_from_json(
            state = state,
            data = selection_data
        )

    @property
    def market(self):
        return self.state.markets.get(
            self.market_id
        )

    @property
    def pool(self):
        if self.market:
            return self.market.pool
        return None

   
    @property
    def positions(self):
        return self.state.positions.get(
            self.id
        )

    @property
    def orders(self):
        
        return list(
            filter(
                lambda o: o.selection_id == self.id,
                self.state.orders.values()
            )
        )

    @property
    def fair_price(self):
        
        return self.price_probability/self.market.book_total if self.market.book_total else self.price_probability

    def __repr__(self):
        return f"<Selection: {self.id}>"

    def __str__(self):
        return f"<Selection: {self.id}>"



"""
Utilities Module
----------------
Various utilities that are useful when interacting with Goated state.
- Price tick bucketing for various price formats
- Conversion tools
- Decimal numeric data encoders
- Payoff Signature helpers (for more information on PayoffSignatures see main documentation)

"""


import json
from decimal import Decimal


class DecimalEncoder(json.JSONEncoder):

    recognized_types = [
        str, int, float, list, dict, bool,
    ]

    def default(self, obj):

        if isinstance(obj, Decimal):
            return f"{obj:.18f}"
            
        elif not type(obj) in self.recognized_types:
            return str(obj)

        return json.JSONEncoder.default(self, obj)


class OrderEncoder(json.JSONEncoder):

    recognized_types = [
        str, int, float, list, dict, bool,
    ]

    def default(self, obj):

        if isinstance(obj, Decimal):
            return f"{obj:.6f}"
            
        elif not type(obj) in self.recognized_types:
            return obj.value

        return json.JSONEncoder.default(self, obj)



from .exceptions import *
import json
from .base_client import BaseClient
from typing import List
from ..instructions import NewOrderArgs, CancelOrderArgs, ReduceOrderArgs, UpdateOrderPersistenceArgs, PriceFormat, OrderStatus, OrderSide, OrderType, QuantityFormat

class TradingClient(BaseClient):


    # EXCHANGE (GENERAL/PUBLIC) INFORMATION

    def get_categories(
        self,
        category_ids: List[int] = []
    ):
        '''
        Returns a list of categories and category meta data.
        category_ids are an optional list of filter parameters which can be passed to narrow the request.
        '''
        filters = {}
        if category_ids != []:
            filters['category_id'] = ','.join([str(s) for s in category_ids])

        response = self._get_objects(
            endpoint = "exchange/categories",
            filters = filters
        )   

        return response

    def get_subcategories(
        self,
        category_ids: List[int] = [],
        subcategory_ids: List[int] = [],
    ):
        '''
        Returns a list of subcategories and subcategory meta data.
        category_ids and subcategory_ids are an optional list of filter parameters which can be passed to narrow the request.
        '''
        filters = {}
        if category_ids != []:
            filters['category_id'] = ','.join([str(s) for s in category_ids])
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])

        response = self._get_objects(
            endpoint = "exchange/subcategories",
            filters = filters
        )   

        return response
    
    def get_events(
        self,
        category_ids: List[int] = [],
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
    ):
        '''
        Returns a list of events and event meta data.
        category_ids, subcategory_ids and event_ids are an optional list of filter parameters which can be passed to narrow the request.
        '''

        filters = {}
        if category_ids != []:
            filters['category_id'] = ','.join([str(s) for s in category_ids])
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
            
        response = self._get_objects(
            endpoint = "exchange/events",
            filters = filters
        )   

        return response



    def get_markets(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        pool_ids: List[int] = [],
        market_ids: List[int] = [],
        market_type_ids: List[int] = [],
        currencies: List[str] = [],
        lines: List[float] = [],
        going_in_play: bool = None,
        main_market: bool = None,
    ):
        '''
        Returns a list of markets and market data for markets which meet the filter criteria.
        The following optional filter parameters can be used to narrow the search:
        - subcategory_ids
        - event_ids
        - pool_ids
        - market_type_ids
        - currencies
        - lines
        - going_in_play
        - main_market
        '''
        filters = {}
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if pool_ids != []:
            filters['pool_id'] = ','.join([str(p) for p in pool_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if lines != []:
            filters['line'] = ','.join([str(l) for l in lines])
        if currencies != []:
            filters['currency'] = ','.join([str(m) for m in currencies])
        if market_type_ids != []:
            filters['market_type'] = ','.join([str(m) for m in market_type_ids])
        if going_in_play != None:
            filters['going_in_play'] = "true" if going_in_play else "false"
        if main_market != None:
            filters['main_market'] = "true" if main_market else "false"

        response = self._get_objects(
            endpoint = "exchange/markets",
            filters = filters
        )   

        return response

    def get_selections(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        pool_ids: List[int] = [],
        market_ids: List[int] = [],
        selection_ids: List[int] = [],
        market_type_ids: List[int] = [],
        currencies: List[str] = [],
        lines: List[float] = [],
        going_in_play: bool = None,
        main_market: bool = None,
    ):
        '''
        Returns a list of selections and selection data for selections which meet the filter criteria.
        The following optional filter parameters can be used to narrow the search:
        - subcategory_ids
        - event_ids
        - pool_ids
        - market_ids
        - selection_ids
        - market_type_ids
        - currencies
        - lines
        - going_in_play
        - main_market
        '''
        filters = {}
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if pool_ids != []:
            filters['pool_id'] = ','.join([str(p) for p in pool_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if selection_ids != []:
            filters['selection_id'] = ','.join([str(s) for s in selection_ids])
        if lines != []:
            filters['line'] = ','.join([str(l) for l in lines])
        if currencies != []:
            filters['currency'] = ','.join([str(m) for m in currencies])
        if market_type_ids != []:
            filters['market_type'] = ','.join([str(m) for m in market_type_ids])
        if going_in_play != None:
            filters['going_in_play'] = 1 if going_in_play else 0
        if main_market != None:
            filters['main_market'] = 1 if main_market else 0

        response = self._get_objects(
            endpoint = "exchange/selections",
            filters = filters
        )  

        return response

    def get_orderbooks(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        pool_ids: List[int] = [],
        market_ids: List[int] = [],
        selection_ids: List[int] = [],
        main_market: bool = None,
        hybrid: bool = True,
        max_depth: bool = None,
        price_format: PriceFormat = None,
        quantity_format: QuantityFormat = None,
    ):
        '''
        Returns orderbook data for selections which meet the filter criteria.
        The following optional filter parameters can be used to narrow the search:
        - subcategory_ids
        - event_ids
        - pool_ids
        - market_ids
        - selection_ids
        - main_market
        The following parameters can be used to adjust the data which is returned:
        - hybrid: Defaults to True. If True will return the hybrid orderbook view - both 'real' and 'synthetic' or 'cross-matched' orders.
        - max_depth: The maximum number of price ticks to return on any given selection and side. Defaults to provide all.
        - price_format: The format of the price to be returned. Defaults to Probability. Decimal and American odds formats are possible.
        - quantity_format: The format of the quantity to be returned. Defaults to Payout. Stake format also possible.
        '''
        filters = {}
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if pool_ids != []:
            filters['pool_id'] = ','.join([str(p) for p in pool_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if selection_ids != []:
            filters['selection_id'] = ','.join([str(s) for s in selection_ids])
        if main_market != None:
            filters['main_market'] = "true" if main_market else "false"
        if hybrid != None:
            filters['hybrid'] = "true" if hybrid else "false"
        if max_depth != None:
            filters['max_depth'] = str(max_depth)
        if price_format != None:
            filters['price_format'] = str(price_format)
        if quantity_format != None:
            filters['quantity_format'] = str(quantity_format)

        response = self._get_objects(
            endpoint = "exchange/orderbook",
            filters = filters
        )  

        return response

    def get_price_histories(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        pool_ids: List[int] = [],
        market_ids: List[int] = [],
        selection_ids: List[int] = [],
        main_market: bool = None,
        interval: int = None, # 300, 900, 3600, 14400, 28800, 86400
        price_format: PriceFormat = None,
        include_components: bool = None,
    ):
        '''
        Returns price history for selections which meet the filter criteria.
        The following optional filter parameters can be used to narrow the search:
        - subcategory_ids
        - event_ids
        - pool_ids
        - market_ids
        - selection_ids
        - main_market
        The following parameters can be used to adjust the data which is returned:
        - interval: The interval of timesteps at which prices are provided, in seconds. Possible values: 300, 900, 3600, 14400, 28800, 86400. By default, the server will return the interval which approximately 30 time steps.
        - price_format: The format of the price to be returned. Defaults to Probability. Decimal and American odds formats are possible.
        - include_components: Defaults to False. If True, returns a breakdown of the best bid, best ask and last traded price at each time step.
        '''
        filters = {}
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if pool_ids != []:
            filters['pool_id'] = ','.join([str(p) for p in pool_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if selection_ids != []:
            filters['selection_id'] = ','.join([str(s) for s in selection_ids])
        if main_market != None:
            filters['main_market'] = "true" if main_market else "false"
        if include_components != None:
            filters['include_components'] = "true" if include_components else "false"
        if price_format != None:
            filters['price_format'] = str(price_format)
        if interval != None:
            filters['interval'] = str(interval)

        response = self._get_objects(
            endpoint = "exchange/history",
            filters = filters
        )  

        return response


    

    # EXCHANGE USER-SPECIFIC INFORMATION

    def get_balances(
        self,
    ):
        '''
        Returns the balances (for each applicable supported currency) held by the authenticated user account.
        '''
        response = self._send_get(
            endpoint = "balances",
        )   
        return response

    def get_orders(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        market_ids: List[int] = [],
        selection_ids: List[int] = [],
        order_ids: List[int] = [],
        status: List[OrderStatus] = []
    ):
        '''
        Gets orders owned by the authenticaed user account.
        By default this query will return orders which have some non-zero 'open' or unmatched component. i.e. orders which have
        been fully cancelled, fully matched or for which trading has ended will not appear in the response.
        Optional filter parameters:
        - subcategory_ids
        - event_ids 
        - market_ids
        - selection_ids
        - order_ids
        - status
        '''
        filters = {}
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if selection_ids != []:
            filters['selection_id'] = ','.join([str(s) for s in selection_ids])
        if order_ids != []:
            filters['order_id'] = ','.join([str(o) for o in order_ids])
        if status != []:
            filters['status'] = ','.join([str(s) for s in selection_ids])
            
        response = self._get_objects(
            endpoint = "exchange/orders",
            filters = filters
        )

        return response

    def get_positions(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        market_ids: List[int] = [],
        selection_ids: List[int] = [],
    ):
        '''
        Gets matched positions currently held by the authenticaed user account.
        This provides the net/aggregate payout and stake for any selections (within the filtered subset)
        in active markets, for which the user has a non-zero matched exposure.
        Further breakdown is provided by selection, side and price tick.
        Optional parameters:
        - subcategory_ids
        - event_ids 
        - market_ids
        - selection_ids
        '''
        filters = {}
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if market_ids != []:
            filters['market_id'] = ','.join([str(m) for m in market_ids])
        if selection_ids != []:
            filters['selection_id'] = ','.join([str(s) for s in selection_ids])

        response = self._get_objects(
            endpoint = "exchange/positions",
            filters = filters
        )   

        return response

    def get_exposures(
        self,
        subcategory_ids: List[int] = [],
        event_ids: List[int] = [],
        pool_ids: List[int] = [],
    ):
        '''
        Gets the authenticated user's risk exposure profile for a given Pool.
        (Pools are at the level of related markets where some risk offsetting between related markets
        is allowed by the Goated risk engine.)
        Optional parameters:
        - subcategory_ids
        - event_ids 
        - market_ids
        '''
        filters = {}
        if event_ids != []:
            filters['event_id'] = ','.join([str(e) for e in event_ids])
        if subcategory_ids != []:
            filters['subcategory_id'] = ','.join([str(s) for s in subcategory_ids])
        if pool_ids != []:
            filters['pool_id'] = ','.join([str(m) for m in pool_ids])
     

        response = self._get_objects(
            endpoint = "exchange/exposures",
            filters = filters
        )   

        return response


    # EXCHANGE ACTIONS

    def place_orders(
        self,
        orders: List[NewOrderArgs],
        asynchronous: bool = False # TODO: Functionality not yet supported
    ): 
        '''
        Places a list of orders on behalf of the authenticated user.
        Parameters:
        - orders: A list of valid NewOrderArgs
        - asynchronous: (Not currently supported) indicating whether or not the server should hang and await completion of the
        matching engine's round trip before providing an API response, or return a tentative confirmation once the orders have been
        placed.
        '''
        if type(orders) != list:
            orders = [orders]
        data = [
            o.serialize_to_dict()
            for o in orders
        ]
        print(data)
        response = self._send_post(
            endpoint = "exchange/orders/new",
            data = data,
            params = {}
        )
        return response

    def cancel_orders(
        self,
        orders: List[CancelOrderArgs],
        asynchronous: bool = False # TODO: Functionality not yet supported
    ): 
        '''
        Requests to cancel existing orders on behalf of the authenticated user.
        Parameters:
        - orders: A list of valid CancelOrderArgs
        - asynchronous: (Not currently supported) indicating whether or not the server should hang and await completion of the
        matching engine's round trip before providing an API response, or return a tentative confirmation once the cancel 
        order request have been placed.
        '''
        if type(orders) != list:
            orders = [orders]
        data = [
            o.serialize_to_dict()
            for o in orders
        ]
        response = self._send_post(
            endpoint = "exchange/orders/cancel",
            data = data,
            params = {}
        )
        return response

    def reduce_orders(
        self,
        orders: List[ReduceOrderArgs],
        asynchronous: bool = False # TODO: Functionality not yet supported
    ): 
        '''
        Requests to reduce the quantity of existing orders on behalf of the authenticated user.
        Parameters:
        - orders: A list of valid ReduceOrderArgs
        - asynchronous: (Not currently supported) indicating whether or not the server should hang and await completion of the
        matching engine's round trip before providing an API response, or return a tentative confirmation once the reduce 
        order request have been placed.
        '''
        if type(orders) != list:
            orders = [orders]
        data = [
            o.serialize_to_dict()
            for o in orders
        ]
        response = self._send_post(
            endpoint = "exchange/orders/edit",
            data = data,
            params = {}
        )
        return response

    def update_persistence_orders(
        self,
        orders: List[UpdateOrderPersistenceArgs],
        asynchronous: bool = False # TODO: Functionality not yet supported
    ): 
        '''
        Requests to update the persistence type for existing orders on behalf of the authenticated user.
        Parameters:
        - orders: A list of valid CancelOrderArgs
        - asynchronous: (Not currently supported) indicating whether or not the server should hang and await completion of the
        matching engine's round trip before providing an API response, or return a tentative confirmation once the persistence type 
        update requests have been placed.
        '''
        if type(orders) != list:
            orders = [orders]
        data = [
            o.serialize_to_dict()
            for o in orders
        ]
        response = self._send_post(
            endpoint = "exchange/orders/edit",
            data = data,
            params = {}
        )
        return response


from decimal import Decimal
from typing import List
import json
import time

class Position():
    
    def __init__(
        self,
        probability_price: Decimal,
        decimal_price: Decimal,
        american_price: Decimal,
        payout: Decimal,
        stake: Decimal,
    ):
        self.probability_price = probability_price
        self.decimal_price = decimal_price
        self.american_price = american_price
        self.payout = payout
        self.stake = stake


class SelectionPosition():

    def __init__(
        self,
        state,
        market_id: int,
        selection_id: int,
        net_payout: Decimal,
        net_stake: Decimal,
        buys: List[Position],
        sells: List[Position],
    ):
        self.state = state
        self.market_id = market_id
        self.selection_id = selection_id
        self.net_payout = net_payout
        self.net_stake = net_stake
        self.buys = buys
        self.sells = sells
        self._update_timestamp = time.time()

    @property
    def probability_price(
        self,
    ):
        return self.net_stake / self.net_payout if self.net_payout != Decimal(0) else None

    @classmethod
    def load_from_json(
        cls,
        state,
        data: str,
    ):
        if type(data) == str:
            data = json.loads(data)
        return cls(
            state = state,
            market_id = data.get('market'),
            selection_id = data.get('id'),
            net_stake = Decimal(data.get('net_stake')),
            net_payout = Decimal(data.get('net_payout')),
            buys = [
                Position(
                    probability_price = Decimal(p.get('probability_price')),
                    decimal_price = Decimal(p.get('decimal_price')),
                    american_price = Decimal(p.get('american_price')),
                    payout = Decimal(p.get('payout')),
                    stake = Decimal(p.get('stake'))
                )
                for p in data.get('buys')
            ],
            sells = [
                Position(
                    probability_price = Decimal(p.get('probability_price')),
                    decimal_price = Decimal(p.get('decimal_price')),
                    american_price = Decimal(p.get('american_price')),
                    payout = Decimal(p.get('payout')),
                    stake = Decimal(p.get('stake'))
                )
                for p in data.get('sells')
            ]
        )


    @property
    def selection(self):
        return self.state.selections.get(
            self.selection_id
        )
    
    @property
    def market(self):
        if self.selection:
            return self.selection.market
        return None

    @property
    def pool(self):
        if self.market:
            return self.market.pool
        return None

    
    def __repr__(self):
        return f"<Position: {self.market_id} {self.selection_id} | S: {self.net_stake:.2f} / P: {self.net_payout:.3f} >"

    def __str__(self):
        return f"<Position: {self.market_id} {self.selection_id} | S: {self.net_stake:.2f} / P: {self.net_payout:.3f} >"




from ..instructions import PriceFormat
from decimal import Decimal

def convert_probability_to_american(probability_price: Decimal):
    if probability_price == None or Decimal(probability_price) >= Decimal(1) or Decimal(probability_price) <= Decimal(0):
        return 0
    if probability_price > 0.5:
        return -(probability_price / (1 - probability_price)) * 100
    else:
        return ((1 - probability_price) / probability_price) * 100



def convert_to_probability(
    price: Decimal,
    price_format: PriceFormat
):
    if type(price) != Decimal:
        price = Decimal(price)

    if price_format == PriceFormat.PROBABILITY:
        return price
    elif price_format == PriceFormat.DECIMAL:
        return Decimal(1) / price if price != Decimal(0) else Decimal(1)
    elif price_format == PriceFormat.AMERICAN:
        if price > 0:
            return Decimal(100) / (price + Decimal(100))
        else:
            return (Decimal(-1) * price)/ ( (Decimal(-1) * price) + Decimal(100) )
    else:
        return None




import numpy as np
from typing import List, Union
from decimal import Decimal


class InvalidPayoffSignature(Exception):

    def __init__(self, payoff_signature):
        self.message = f'InvalidPayoffSignature: Invalid payoff signature: {payoff_signature}'
        super().__init__(self.message)


def check_only_valid_values(
    payoff_string: str,
):
    letters = set(payoff_string)
    letters_allowed = ["L" ,"R", "W", "H", "K"]
   
    for letter in letters:
        if letter not in letters_allowed:
            raise InvalidPayoffSignature(payoff_string)
    
    return True

# def check_for_price_dependent_payoffs(
#     payoff_string: str,
# ):
#     letters = set(payoff_string)
#     if "R" in  letters or "H" in letters or "K" in letters:
#         return True
#     return False

class PayoffSignature():

    def __init__(
        self,
        payoff_string: str,
        dimensions: int = 2,
    ):
        '''

        '''
        if dimensions not in [1, 2]:
            raise InvalidPayoffSignature(payoff_string)

        letters = set(payoff_string)
        letters_allowed = ["L" ,"R", "W", "H", "K"]

        for letter in letters:
            if letter not in letters_allowed:
                raise InvalidPayoffSignature(payoff_string)

        self._dimensions = dimensions
        self._string = payoff_string.upper()
        self._length = len(self._string)

        self._has_refund = True if "R" in letters else False
        self._has_half_win = True if "H" in letters else False
        self._has_half_loss = True if "K" in letters else False
        self._has_price_dependent_payoff = self._has_refund or self._has_half_loss or self._has_half_win

        # Where there is a price dependency (i.e. it's possible for get a refund or half-refund)
        #  we may need to do additional checks when optimizing for match payoffs


    def to_string(
        self
    ):
        return self._string

    @classmethod
    def from_string(
        cls,
        payoff_string: str,
        dimensions: int = 2
    ):
        return cls(
           payoff_string = payoff_string,
           dimensions = dimensions 
        )


    @classmethod
    def empty_exposure_vector(
        cls,
        length: int,
        dimensions: int = 2,
    ):
        return np.asarray([Decimal(0) for i in range(length)])


    def to_grid_representation(
        self
    ) -> str:
        '''
        Returns a grid representation where:
          Dark cells =  Win of Half Win
          Shaded cells =  Refund
          Empty cells =  Loss of Half-Loss
        '''
        size = int(np.sqrt(self._length))
        divider = '-'*((size+1)*4+1) + '\n'
        string = divider
        string += '|   '
        for a in range(size):
            string += f"|{a:2d} "
        string += '|\n'
        for h in range(size):
            string += f"|{h:2d} "
            for a in range(size):
                value = self._string[h*(size) + a]
                string += "|▉▉▉" if value in ['W', 'H'] else "|   " if value in ["R"] else "|░░░"
            string += '|\n'
        string += divider
        return string
        
        
    def invert(
        self
    ) -> super:
        '''
        Returns a new instance of PayoffSignture where the payoff is the
        inverse of self.
        '''
        inverted_string = self._string.replace('W', 'l').replace('L', 'w').replace('H', 'k').replace('K','h').upper()
        return PayoffSignature(
            payoff_string = inverted_string,
            dimensions = self._dimensions
        )

    def generate_payoff_vector(
        self,
        price: Decimal # This is the 'post price' - i.e. for a Buyer, the actual price, for a Seller (1-actual price)
    ): 
        price = Decimal(price)
        values = []
        for l in self._string:
            if l == "W":
                value = Decimal(1) 
            elif l == "L":
                value = Decimal(0)
            elif l == "R":
                value = Decimal(price)
            elif l == "H":
                value = Decimal(0.5)*(Decimal(1)+price)
            elif l == "K":
                value = Decimal(0.5)*Decimal(price)
            else:
                raise InvalidPayoffSignature(self._string)
            values += [value]    
        vector = np.asarray(values)
        if len(vector) != self._length:
            raise InvalidPayoffSignature(self._string)

        return vector


    def generate_negative_only_net_payoff_vector(
        self,
        price: Decimal # This is the 'post price' - i.e. for a Buyer, the actual price, for a Seller (1-actual price)
    ): 
        # Net meaning for a win, we have 1-price, for a loss -price, etc.
        price = Decimal(price)
        vector = self.generate_net_payoff_vector(price).copy()
        vector[vector >= 0] = Decimal(0)  # i.e. only return values which are <0 else make them 0.
        return vector


    def generate_net_payoff_vector(
        self,
        price: Decimal # This is the 'post price' - i.e. for a Buyer, the actual price, for a Seller (1-actual price)
    ): 
        # Net meaning for a win, we have 1-price, for a loss -price, etc.
        price = Decimal(price)
        vector = self.generate_payoff_vector(price)
        vector = np.subtract(vector, price) 

        return vector


    def generate_payout_per_unit_risk_vector(
        self,
        price: Decimal # This is the 'post price' - i.e. for a Buyer, the actual price, for a Seller (1-actual price)
    ): 
        price = Decimal(price)
        values = []
        for l in self._string:
            if l == "W":
                value = Decimal(1)/price
            elif l == "L":
                value = Decimal(0)
            elif l == "R":
                value = Decimal(1)
            elif l == "H":
                value = Decimal(0.5)/price
            elif l == "K":
                value = Decimal(0.5)
            else:
                raise InvalidPayoffSignature(self._string)
            values += [value]    
        vector = np.asarray(values)
        if len(vector) != self._length:
            raise InvalidPayoffSignature(self._string)
        return vector

    def generate_net_return_per_unit_risk_vector(
        self,
        price: Decimal # This is the 'post price' - i.e. for a Buyer, the actual price, for a Seller (1-actual price)
    ): 
        # i.e. subtract the $1 risk from each to give the net return
        
        price = Decimal(price)
        vector = self.generate_payout_per_unit_risk_vector(price)
        vector = np.subtract(vector, Decimal(1)) 

        return vector


    def get_minimum_values_from_payoff_vector(
        self,
        payoff_vector: np.ndarray,
        payoff_letter: str
    ):
        '''
        Gets the minimum value for any cells of a particular payoff type
        (as indicated by the payoff letter - i.e. "R", "H", etc)
        for a payoff vector provided.
        Returns None if no value of this type is present.
        '''
        payoff_letter = payoff_letter.upper()
        if len(payoff_vector) != self._length:
            raise InvalidPayoffSignature(self._string)
        result = None
        for idx, letter in enumerate(self._string):
            if letter == payoff_letter:
                if result == None or payoff_vector[idx] < result:
                    result = payoff_vector[idx]
        return result.copy()




"""
Goated Package
--------------
The Goated Python package contains wrappers for the Goated platform API and tools for handling local copies of state.
For more information on the Goated exchange and to create or manage your account please visit https://goated.com/

"""

from .version import __version__
from .client import *
from .state import *
from .instructions import *
from .utils import *


import json
from typing import Union, List
from .order import Order
from .position import SelectionPosition
from .pool import Pool
from .market import Market
from .selection import Selection

class Container():

    def __init__(
        self,
    ):
        self.balances = {}
        self.positions = {}
        self.orders = {}
        self.pools = {}
        self.markets = {}
        self.selections = {}

    def update_balances(
        self,
        data: Union[str, dict]
    ):  
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for balance in data:
            self.balances[balance.get('currency')] = balance

    def remove_balance(
        self,
        currency: str
    ):
        if currency in self.balances.keys():
            self.orders.pop(str(currency))
        else:
            print('No balance found for this currency')


    def update_orders(
        self,
        data: Union[dict, str, list]
    ):
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for order_data in data:

            # Create an in-memory representation
            order = Order.load_from_json(
                state = self,
                data = order_data
            )

            # If the order has been removed, pop it from our collection
            if order.is_removed and order.id in self.orders.keys():
                self.orders.pop(order.id)

            # Otherwise, if it's open (not failed/pending) then we should include or update
            #  it within our collection
            elif order_data.get('is_open') and order_data.get('status') not in ['pending', 'failed']:
                self.orders[order_data.get('id')] = order
      
            else:
                pass # Don't store orders which aren't open
            
            
    def remove_order(
        self,
        order_id: int
    ):
        if order_id in self.orders.keys():
            self.orders.pop(int(order_id))
        else:
            print('No order found for this order_id')


    def update_positions(
        self,
        data: Union[dict, str, list]
    ):
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for position_data in data:
            position = SelectionPosition.load_from_json(
                state = self,
                data = position_data
            )
            self.positions[position_data.get('id')] = position
            if position.pool:
                position.pool._positions_updated = True
            


    def remove_position(
        self,
        selection_id: int
    ):
        if selection_id in self.positions.keys():
            self.positions.pop(int(selection_id))
        else:
            print('No positions found for this selection_id')


    def update_pools(
        self,
        data: Union[dict, str, list]
    ):
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for pool_data in data:
            if pool_data.get('id') not in self.pools.keys():
                self.pools[pool_data.get('id')] = Pool.load_from_json(
                    state = self,
                    data = pool_data
                )
            else:
                self.pools[pool_data.get('id')].update_from_json(
                    state = self,
                    data = pool_data
                )

    def remove_pool(
        self,
        pool_id: int,
        remove_markets: bool = True,
        remove_selections: bool = True
    ):
        if pool_id in self.pools.keys():
            pool = self.pools.pop(int(pool_id))
            if remove_markets:
                for market in pool.markets:
                    self.remove_market(
                        market_id = market.id,
                        remove_selections = remove_selections
                    )
        else:
            print('No pool found for this id')
    
    def update_markets(
        self,
        data: Union[dict, str, list]
    ):
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for market_data in data:
            market = Market.load_from_json(
                state = self,
                data = market_data
            )
            self.markets[market.id] = market
            if market.pool:
                market.pool._markets_updated = True
           

    def remove_market(
        self,
        market_id: int,
        remove_selections: bool = True
    ):
        if market_id in self.markets.keys():
            market = self.markets.pop(int(market_id))
            if remove_selections:
                for selection in market.selections:
                    self.remove_selection(
                        selection_id = selection.id
                    )
        else:
            print('No market found for this id')


    def update_selections(
        self,
        data: Union[dict, str, list]
    ):
        if type(data) == str:
            data = json.loads(data)
        if type(data) != list:
            data = [data]
        for selection_data in data:
            selection = Selection.load_from_json(
                state = self,
                data = selection_data
            )
            self.selections[selection_data.get('id')] = selection
            if selection.pool:
                selection.pool._selections_updated = True
            

    def remove_selection(
        self,
        selection_id: int
    ):
        if selection_id in self.selections.keys():
            selection = self.selections.pop(int(selection_id))
        else:
            print('No selection found for this id')

  
    @property
    def market_ids(
        self,
    ): 
        return list(self.markets.keys())
    
    @property
    def selection_ids(
        self,
    ): 
        return list(self.selections.keys())
    
    @property
    def order_ids(
        self,
    ): 
        return list(self.orders.keys())


"""
Instructions Module
-------------------
This module deals primarily with assembling and relaying instructions
to the Goated Exchange on behalf of the authenticated user account.

Instructions currently supported include:
- Requests to place new orders on the exchange
- Requests to cancel or remove (in full) existing orders 
- Requests to reduce the remaining (unmatched) quantity of existing orders
- Requests to update the persistency type of an existing order
"""


from .enums import *
from .new_order import NewOrderArgs
from .cancel_order import CancelOrderArgs
from .reduce_order import ReduceOrderArgs
from .update_persistence import UpdateOrderPersistenceArgs



from decimal import Decimal
from ..order import Order, OrderSide
from typing import List
from ...instructions import NewOrderArgs, CancelOrderArgs, ReduceOrderArgs, OrderSide, OrderType, QuantityFormat, PersistenceType, PriceFormat
from .pool_profile import PoolProfile

class CurrentLiquidityProfile(PoolProfile):


    @classmethod
    def generate_from_orders(
        cls,
        pool,
        orders: List[Order] = None,
    ):
        instance = cls(
            pool = pool
        )

        if orders == None and pool:
            orders = pool.orders

        for order in orders:
            instance.add(
                market_id = int(order.market_id), 
                selection_id = int(order.selection_id),
                side = order.side,
                price = Decimal(order.probability_price),
                quantity = Decimal(order.payout_quantity_remaining)
            )
        return instance


    def solve_instructions_to_target(
        self,
        target_profile, #TargetLiquidityProfile
    ):
        '''
        Returns a dict of 'new', 'cancel' and 'reduce' lists, containing
        NewOrderArg, CancelOrderArgs and ReduceOrderArgs respectively.
        This is the optimal set of instructions to move from the current liquidity
        profile (self) to the target profile (provided).
        [NB: Optimum where we're optimizing for first removing the newest orders,
        rather than for fewest actions.]
        '''
        instructions = {
            'new': [],
            'cancel': [],
            'reduce': []
        }
        
        delta, actions = self.__class__.generate_delta_from_two_profiles(
            starting_profile = self,
            target_profile = target_profile,
        )

        delta.print()

        for action in actions:
            if action['type'] == 'new':
                instructions['new'] += [
                    NewOrderArgs(
                        market_id = action['market_id'],
                        selection_id = action['selection_id'],
                        side = action['side'],
                        type = OrderType.LIMIT,
                        price = Decimal(action['price']),
                        price_format = PriceFormat.PROBABILITY,
                        quantity = Decimal(action['quantity']),
                        quantity_format = QuantityFormat.PAYOUT,
                        persistence_type = PersistenceType.INCLUDE_IN_AUCTION
                    )
                ]
            elif action['type'] == 'cancel':
                relevant_orders = filter(
                    lambda o: (o.market_id == action['market_id'] and o.selection_id == action['selection_id'] and o.side == OrderSide(action['side']) and abs(o.price - action['price']) < Decimal(0.001)),
                    self.pool.orders
                )
                for o in relevant_orders:
                    instructions['cancel'] += [
                        CancelOrderArgs(
                            id = o.order_id
                        )
                    ]
            elif action['type'] == 'reduce':
                reduce_quantity = -action['quantity']
                relevant_orders = filter(
                    lambda o: (o.market_id == action['market_id'] and o.selection_id == action['selection_id'] and o.side == OrderSide(action['side']) and abs(o.price - action['price']) < Decimal(0.001)),
                    self.pool.orders
                )
                relevant_orders = sorted( # sorted by id in reverse (which is a proxy for sequence) so that we leave older orders (greater priority)
                    relevant_orders,
                    key = lambda o: int(o.id),
                    reverse = True
                )
                for o in relevant_orders:
                    if o.payout_quantity_remaining > reduce_quantity:
                        new_quantity = o.payout_quantity_remaining - reduce_quantity
                        reduce_quantity = 0
                    else:
                        new_quantity = 0
                        reduce_quantity -= o.payout_quantity_remaining
                    if new_quantity > 0:
                        instructions['reduce'] += [
                            ReduceOrderArgs(
                                order_id = o.id,
                                new_quantity = new_quantity,
                                quantity_format = QuantityFormat.PAYOUT
                            )
                        ]
                    else:
                        instructions['cancel'] += [
                            CancelOrderArgs(
                                order_id = o.id
                            )
                        ]
                    if reduce_quantity == 0: # Exit when we've reduced by the quantity required
                        break
            else:
                pass

        return instructions


    def solve_instructions_to_clear(
        self,
    ):
        '''
        Returns a dict of 'cancel' ('new' and 'reduce' there too, but empty)
        lists, containing CancelOrderArgs (NewOrderArg, ReduceOrderArg) required
        to clear the entire liquidity profile
        '''
        instructions = {
            'new': [],
            'cancel': [],
            'reduce': []
        }
        for o in self.orders:
            instructions['cancel'] += [
                CancelOrderArgs(
                    id = o.order_id
                )
            ]
        return instructions
        

    def print(
        self
    ):
        print('='*40)
        print('CURRENT LIQUIDITY PROFILE')
        print(f'Pool: {self.pool.id}')
        print('-'*40)
        for market_id in self.data.keys():
            print(f'MARKET: {market_id}')
            for selection_id in self.data[market_id].keys():
                print(f' - SELECTION: {selection_id}')
                for side in self.data[market_id][selection_id].keys():
                    print(f'   - {side.value.upper()}')
                    for price, quantity in sorted(self.data[market_id][selection_id][side].items()):
                        print(f'      {price:.3f}: {quantity:8.2f}')
            print('-'*40)
        print('='*40)







from .pool_profile import PoolProfile


class TargetLiquidityProfile(PoolProfile):

    # Inherits all necessary functionality 
    # from the parent PoolProfile class

    def print(
        self
    ):
        print('='*40)
        print('TARGET LIQUIDITY PROFILE')
        print(f'Pool: {self.pool.id}')
        print('-'*40)
        for market_id in self.data.keys():
            print(f'MARKET: {market_id}')
            for selection_id in self.data[market_id].keys():
                print(f' - SELECTION: {selection_id}')
                for side in self.data[market_id][selection_id].keys():
                    print(f'   - {side.value.upper()}')
                    for price, quantity in sorted(self.data[market_id][selection_id][side].items()):
                        print(f'      {price:.3f}: {quantity:8.2f}')
            print('-'*40)
        print('='*40)


from typing import List
from decimal import Decimal
from ..instructions import PriceFormat

PRICE_ROUNDING_TOLERANCE = Decimal(10**-5)

class PriceSchema():

    def __init__(
        self,
        price_format: PriceFormat,
        absolute_min: Decimal,
        absolute_max: Decimal,
        mins: List[Decimal],
        maxs: List[Decimal],
        increments: List[Decimal]
    ):
        self._price_format = price_format
        self._format_absolute_min = absolute_min
        self._format_absolute_max = absolute_max
        self._format_mins = mins
        self._format_maxs = maxs
        self._format_increments = increments
        self.get_min_max()
        self.generate_range()


    def get_min_max(
        self
    ):
        if self._price_format == PriceFormat.PROBABILITY:
            self.max = Decimal(self._format_maxs[-1])
            self.min = Decimal(self._format_mins[0])
        elif self._price_format == PriceFormat.DECIMAL:
            self.max = Decimal(1/self._format_mins[0])
            self.min = Decimal(1/self._format_maxs[-1])
        elif self._price_format == PriceFormat.AMERICAN:
            self.max = Decimal(self._format_mins[0]/(100+self._format_mins[0]) if self._format_mins[0] < 0 else 100/(100+self._format_mins[0]))
            self.min = Decimal(self._format_maxs[-1]/(100+self._format_maxs[-1]) if self._format_maxs[-1] < 0 else 100/(100+self._format_maxs[-1]))

    @classmethod
    def load(
        cls,
        config: dict
    ):
        return cls(
            price_format = PriceFormat(config['price_format']),
            absolute_min = config['absolute_min'],
            absolute_max = config['absolute_max'],
            mins = config['mins'],
            maxs = config['maxs'],
            increments= config['increments'], 
        )

    def convert_from_probability(
        self,
        probability_price: Decimal
    ):
        if self._price_format == PriceFormat.DECIMAL:
            format_price = Decimal(1 / probability_price)
        elif self._price_format == PriceFormat.AMERICAN:
            if probability_price < 0.5:
                format_price = Decimal(Decimal(1/probability_price)*100 - 100)
            else:
                format_price = Decimal((probability_price*100)/(1-probability_price))
        else:
            format_price = probability_price
        return format_price


    def convert_to_probability(
        self,
        format_price: Decimal
    ):
        if self._price_format == PriceFormat.DECIMAL:
            probability_price = Decimal(1/format_price)
            
        elif self._price_format == PriceFormat.AMERICAN:
            if format_price < 0:
                if format_price == Decimal(-100):
                    probability_price = Decimal(0.5)
                else:
                    probability_price = Decimal(format_price/(100+format_price))
            else:
                probability_price = Decimal(100/(100+format_price))

        else:
            probability_price = format_price

        return probability_price 
    
    @property
    def prudent_is_lower(
        self,
    ):
        # Peudent here means a higher probability price
        # i.e. we're rounding the price at which a match is 'sold' up 
        # to the next acceptable tick.
        if self._price_format == PriceFormat.DECIMAL:
            prudent_is_lower = True
        elif self._price_format == PriceFormat.AMERICAN:
            prudent_is_lower = True
        else:
            prudent_is_lower = False
        return prudent_is_lower


    def bucket(
        self,
        raw_price: Decimal,
        bucket_inverted: bool = False,
        prudent: bool = False
    ) -> Decimal:

        # If we're bucketing inverted then operate on price = (1-price)
        # and convert back at the end
        # Prudent direction is the opposite if inverted
        if bucket_inverted:
            raw_price = (1-raw_price)
            prudent_is_lower = not self.prudent_is_lower
        else:
            prudent_is_lower = self.prudent_is_lower

        # If price out of bounds, then we have a problem
        if raw_price <= 0 or raw_price >= 1:
            raise Exception(f'Price is out of bounds: {raw_price}')


        # If the price is within bounds but 
        # outside the acceptable min/max then we can simply
        # return this absolute min/max bound
        if bucket_inverted:
            if raw_price >= Decimal(1) - self.min:
                return self.min
            elif raw_price <= Decimal(1) - self.max:
                return self.max
        else:
            if raw_price >= self.max:
                return self.max
            elif raw_price <= self.min:
                return self.min

        # Convert the probabity price to te format price
        format_raw_price = self.convert_from_probability(
            probability_price = raw_price
        )

        # Find the containing band
        lower_bound_index = next(
            (
                idx for idx, x in enumerate(self._format_mins)
                if x >= format_raw_price
            ),
            None
        )
        lower_bound_index = (
            len(self._format_mins) - 1
            if lower_bound_index == None
            else lower_bound_index - 1
        )
        increment = Decimal(
            self._format_increments[lower_bound_index]
        )

        # Naively round to the nearest tick
        bucketed_format_price = Decimal(
            Decimal(round(format_raw_price / increment)) * increment
        )

        # If prudent, then find the nearest prudent tick
        if prudent:
            if abs(bucketed_format_price - format_raw_price) < Decimal(PRICE_ROUNDING_TOLERANCE):
                pass # Close enough, don't do anything
            elif prudent_is_lower and format_raw_price > bucketed_format_price:
                pass # Already prudent
            elif not prudent_is_lower and format_raw_price < bucketed_format_price:
                pass # Already prudent
            elif prudent_is_lower:
                bucketed_format_price -= increment
            elif not prudent_is_lower:
                bucketed_format_price += increment
                
        # Convert back to probability
        bucketed_probability_price = self.convert_to_probability(
            format_price = bucketed_format_price
        )

        # Flip it back if inverted
        if bucket_inverted:
            bucketed_probability_price = Decimal(1) - bucketed_probability_price

        return bucketed_probability_price

    
    def index_of_nearest_tick_greater_than(
        self,
        probability_price
    ):
        return next(
            (
                index for index, item in enumerate(self.probability_range)
                if item > probability_price
            ),
            None
        )
    
    def index_of_nearest_tick_less_than_equal(
        self,
        probability_price
    ):
        return self.index_of_nearest_tick_greater_than(probability_price) - 1


    def nearest_tick_greater_than(
        self,
        probability_price
    ):
        return self.probability_range(self.index_of_nearest_tick_greater_than(probability_price))
    

    def nearest_tick_less_than_equal(
        self,
        probability_price
    ):
        return self.probability_range(
            self.index_of_nearest_tick_less_than_equal(probability_price)
        )
        
        
    def nearest_tick_with_number_ticks_offset(
        self,
        probability_price,
        ticks_to_offset: int # +/- number of ticks to step away from nearest
    ):
        index_nearest_over = self.index_of_nearest_tick_greater_than(probability_price)
        index_nearest_under= index_nearest_over - 1
        above = Decimal(self.probability_range[index_nearest_over])
        below = Decimal(self.probability_range[index_nearest_under])
        if abs(above - probability_price) < abs(below - probability_price):
            nearest_index = index_nearest_over
        else:
            nearest_index = index_nearest_under
        return self.probability_range[nearest_index + ticks_to_offset]

    def nearest_tick(
        self,
        probability_price
    ):
        return self.nearest_tick_with_number_ticks_offset(
            probability_price = probability_price,
            ticks_to_offset = 0
        )

    def generate_range(
        self,
    ):
        values_in_format = [self._format_absolute_min]
        for idx in range(len(self._format_mins)):
            if self._format_mins[idx] in values_in_format:
                v = self._format_mins[idx] + self._format_increments[idx]
            else:
                v = self._format_mins[idx]
            while v <= self._format_maxs[idx]:
                values_in_format.append(v)
                v += self._format_increments[idx]
            if self._format_maxs[idx] not in values_in_format:
                values_in_format.append(self._format_maxs[idx])
        if self._format_absolute_max not in values_in_format:
            values_in_format.append(self._format_absolute_max)
        
        if self._price_format == PriceFormat.PROBABILITY: # Prices ascending for probability already
            self.format_range = sorted(set(values_in_format))
            self.probability_range = self.format_range
        else:
            self.format_range = sorted(set(values_in_format))[::-1] # Flip prices from descending to ascending for decimal or american format
            self.probability_range = [self.convert_to_probability(format_price=fp) for fp in self.format_range]
        

### PRECONFIGURATIONS ###

DEFAULT_PROBABILITY_SCHEMA_CONFIG = {
    'price_format': 'probability', 
    'absolute_min': 0.001,
    'absolute_max': 0.99,
    'mins': [
        0.001,
        0.002,
        0.005,
        0.01,
        0.02,
        0.05,
        0.1
    ],
    'maxs': [
        0.002,
        0.005,
        0.01,
        0.02,
        0.05,
        0.1,
        0.99
    ],
    'increments': [
        0.0001,
        0.00025,
        0.0005,
        0.001,
        0.0025,
        0.005,
        0.01
    ]
}

DEFAULT_PROBABILITY_SCHEMA = PriceSchema.load(
    config = DEFAULT_PROBABILITY_SCHEMA_CONFIG
)

DEFAULT_DECIMAL_SCHEMA_CONFIG = {
    'price_format': 'decimal',
    'absolute_min': 1.01,
    'absolute_max': 1000.0,
    'mins': [
        1.01,
        2.0,
        3.0,
        4.0,
        6.0,
        10.0,
        20.0,
        30.0,
        50.0,
        100.0
    ],
    'maxs': [
        2.0,
        3.0,
        4.0,
        6.0,
        10.0,
        20.0,
        30.0,
        50.0,
        100.0,
        1000.0
    ],
    'increments': [
        0.01,
        0.02,
        0.05,
        0.1,
        0.2,
        0.5,
        1.0,
        2.0,
        5.0,
        10.0
    ]
}

DEFAULT_DECIMAL_SCHEMA = PriceSchema.load(
    config = DEFAULT_DECIMAL_SCHEMA_CONFIG
)

DEFAULT_AMERICAN_SCHEMA_CONFIG = {
    'price_format': 'american',
    'absolute_min': -100000,
    'absolute_max': 100000,
    'mins': [
        -100000,
        -10000,
        -5000,
        -3000,
        -2000,
        -1000,
        -500,
        -300,
        -200,
        -100,
        100,
        200,
        300,
        500,
        1000,
        2000,
        3000,
        5000,
        10000
    ],
    'maxs': [
        -10000,
        -5000,
        -3000,
        -2000,
        -1000,
        -500,
        -300,
        -200,
        -100,
        100,
        200,
        300,
        500,
        1000,
        2000,
        3000,
        5000,
        10000,
        100000
    ],
    'increments': [
        1000,
        500,
        250,
        100,
        50,
        25,
        10,
        5,
        1,
        1,
        1,
        5,
        10,
        25,
        50,
        100,
        250,
        500,
        1000
    ]
}


DEFAULT_AMERICAN_SCHEMA = PriceSchema.load(
    DEFAULT_AMERICAN_SCHEMA_CONFIG
)


class NoTokenProvided(Exception):
    def __init__(self,):
        self.message = f'No authentication token was provided.'
        super().__init__(self.message)

class GoatedApiException(Exception):
    def __init__(self, response):
        self.message = f"The server returned an error. \n {response} "
        super().__init__(self.message)



"""
State Module
------------
This module is intended to support local state management of 
Goated information.

The module include native Python classes for Goated objects - such as Pools,
Markets, Selections, Orders and Positions.

It also included Profiles tools, which can be used to help model
the liquidity shown by traders or market makers on the exchange.

The Container object is a useful implementation for storing and maintaining
relationships between different objects cached in local memory. i.e. it enables
an implementation to retain and refresh cached information from the Goated API and interact
with the related objects naturally.

"""

from .container import Container
from .pool import Pool
from .market import Market
from .selection import Selection
from .order import Order
from .position import Position
from .profiles import CurrentLiquidityProfile, PoolProfile, MatchedPositionProfile, TargetLiquidityProfile



from goated.client import TradingClient
from goated.state import Container
import os
import json


# =======================================================================
#   Initialize an TradingClient from env variables
# -----------------------------------------------------------------------


client = TradingClient.create_with_api_key(
    url = "https://api.goated.com",
    api_key = "YOUR_GOATED_API_KEY",
    api_secret = "YOUR_GOATED_API_SECRET"
)

print(client)

# Initialize an empty state container
# (used to store state and keep relationships between objects)
state = Container()
print(state)


# =======================================================================
#   Get balances, positions, orders
# -----------------------------------------------------------------------

# Get balances response from API
balances_response = client.get_balances()
# Add/update it within the state container
state.update_balances(balances_response)

# Get positions response from API
positions_response = client.get_positions() # Without any filters
# Add/update it within the state container
state.update_positions(positions_response)

# Get orders response from API
orders_response = client.get_orders() # Without any filters
# Add/update it within the state container
state.update_orders(orders_response)

print(state.__dict__)


# =======================================================================
#   Get categories, subcategories, market types, etc
#     i.e. information that is purely 'listing' rather than dynamic or
#     requiring computation/calcs/transformations
# -----------------------------------------------------------------------

# Get all categories
categories = client.get_categories()  # Without any filters

print('Categories:')
print(categories)

# Get subcategory record for first category
category_id = categories[0].get('id')
subcategories = client.get_subcategories(
    category_ids=[category_id]
) if category_id != None else []
print('Subcategories:')
print(subcategories)

# Get events in the first subcategory
subcategory_id = subcategories[0].get('id') if len(subcategories) > 0 else None
events = client.get_events(
    subcategory_ids=[subcategory_id]
) if subcategory_id != None else []
print('Events:')
print(events)


# Get all markets in the first event
event_id = events[0].get('id') if len(events) > 0 else None
markets = client.get_markets(
    event_id=[event_id]
) if event_id != None else []
print('Markets:')
print(markets)


# Load a pool, market, etc into state container


from .enums import *
from decimal import Decimal
import json
from datetime import datetime
from typing import List
from ..utils.encoder import OrderEncoder


class NewOrderArgs():

    def __init__(
        self,
        market_id: int,
        selection_id: int,
        side: OrderSide,
        type: OrderType,
        price: Decimal,
        price_format: PriceFormat,
        quantity: Decimal,
        quantity_format: QuantityFormat,
        persistence_type: PersistenceType,
        custom_order_id: str = None,
        custom_strategy_id: str = None,
    ):
        self.market = market_id
        self.selection = selection_id
        self.side = side
        self.type = type
        self.price = price
        self.price_format = price_format
        self.quantity = quantity
        self.quantity_format = quantity_format
        self.persistence_type = persistence_type
        self.custom_order_id = custom_order_id
        self.custom_strategy_id = custom_strategy_id


    @classmethod
    def load_from_json(
        self,
        data: str,
    ):
        if type(data) == str:
            data = json.loads(data)
        self.market_id = data.get('market')
        self.selection_id = data.get('selection')
        self.side = OrderSide(data.get('side'))
        self.type = OrderType(data.get('type'))
        self.price = Decimal(data.get('price'))
        self.price_format = PriceFormat(data.get('price_format'))
        self.quantity = Decimal(data.get('quantity'))
        self.quantity_format = QuantityFormat(data.get('quantity_format'))
        self.persistence_type = PersistenceType(data.get('persistence_type'))

    def serialize_to_json(
        self,
    ):
        return json.dumps(
            self.__dict__,
            cls=OrderEncoder
        )
    
    def serialize_to_dict(
        self,
    ):
        d = json.loads(
            self.serialize_to_json()
        )
        keys = [k for k in d.keys()]
        for k in keys:
            if d.get(k) == None:
                d.pop(k)
        return d



from decimal import Decimal
from typing import List
from datetime import datetime
from ..instructions import OrderSide, OrderType, PriceFormat, QuantityFormat, PersistenceType
import json
import time

class Order():
    
    def __init__(
        self,
        state,
        id: int,
        market_id: int,
        selection_id: int,
        currency: str,
        side: OrderSide,
        type: OrderType,
        custom_order_id: str,
        custom_strategy_id: str,
        probability_price: Decimal,
        price: Decimal,
        price_format: PriceFormat,
        quantity: Decimal,
        quantity_format: QuantityFormat,
        persistence_type: PersistenceType,
        status: str,
        payout_quantity_requested: Decimal,
        stake_quantity_requested: Decimal,
        payout_quantity_filled: Decimal,
        stake_quantity_filled: Decimal,
        payout_quantity_cancelled: Decimal,
        stake_quantity_cancelled: Decimal,
        payout_quantity_lapsed: Decimal,
        stake_quantity_lapsed: Decimal,
        payout_quantity_voided: Decimal,
        stake_quantity_voided: Decimal,
        payout_quantity_remaining: Decimal,
        stake_quantity_remaining: Decimal,
        average_matched_price: Decimal,
        removal_reason: str,
        matched_timestamp: datetime,
        lapse_timestamp: datetime,
        cancel_timestamp: datetime,
        created: datetime,
        is_removed = bool,
        is_open  = bool,
        is_matched = bool, 
        is_settled = bool,
    ):
        self.state = state
        self.id = id
        self.market_id = market_id
        self.selection_id = selection_id
        self.currency = currency
        self.side = side
        self.type = type
        self.custom_order_id = custom_order_id
        self.custom_strategy_id = custom_strategy_id
        self.probability_price = probability_price
        self.price = price
        self.price_format = price_format
        self.quantity = quantity
        self.quantity_format = quantity_format
        self.persistence_type = persistence_type
        self.status = status
        self.payout_quantity_requested = payout_quantity_requested
        self.stake_quantity_requested = stake_quantity_requested
        self.payout_quantity_filled = payout_quantity_filled
        self.stake_quantity_filled = stake_quantity_filled
        self.payout_quantity_cancelled = payout_quantity_cancelled
        self.stake_quantity_cancelled = stake_quantity_cancelled
        self.payout_quantity_lapsed = payout_quantity_lapsed
        self.stake_quantity_lapsed = stake_quantity_lapsed
        self.payout_quantity_voided = payout_quantity_voided
        self.stake_quantity_voided = stake_quantity_voided
        self.payout_quantity_remaining = payout_quantity_remaining
        self.stake_quantity_remaining = stake_quantity_remaining
        self.average_matched_price = average_matched_price
        self.removal_reason = removal_reason
        self.matched_timestamp = matched_timestamp
        self.lapse_timestamp = lapse_timestamp
        self.cancel_timestamp = cancel_timestamp
        self.created = created
        self.is_removed= is_removed
        self.is_open = is_open 
        self.is_matched = is_matched 
        self.is_settled= is_settled
        self._pool_id = None
        self._update_timestamp = time.time()

    

    def __repr__(self):
        return f"<Order: {self.id} | {self.selection_id} | {str(self.side)}" + (f"{self.payout_quantity_remaining:.2f} @ {self.probability_price:.3f}") if self.status != 'pending' else "" + f" {self.status.upper()}>"

    def __str__(self):
        return f"<Order: {self.id} | {self.selection_id} | {str(self.side)}" + (f"{self.payout_quantity_remaining:.2f} @ {self.probability_price:.3f}") if self.status != 'pending' else "" + f" {self.status.upper()}>"

    @classmethod
    def load_from_json(
        cls,
        state,
        data: str,
    ):
        if type(data) == str:
            data = json.loads(data)
        return cls(
            state = state,
            id = data.get('id'),
            market_id = data.get('market'),
            selection_id = data.get('selection'),
            currency = data.get('currency'),
            side = OrderSide(data.get('side')),
            type = OrderType(data.get('type')),
            custom_order_id = data.get('custom_order_id'),
            custom_strategy_id = data.get('custom_strategy_id'),
            probability_price = Decimal(data.get('probability_price')) if data.get('probability_price') != None else None,
            price = Decimal(data.get('price')) if data.get('price') != None else None,
            price_format = PriceFormat(data.get('price_format')),
            quantity = Decimal(data.get('quantity')) if data.get('quantity') != None else None,
            quantity_format = QuantityFormat(data.get('quantity_format')),
            persistence_type = PersistenceType(data.get('persistence_type')),
            status = data.get('status'),
            payout_quantity_requested = Decimal(data.get('payout_quantity_requested')) if data.get('payout_quantity_requested') != None else None,
            stake_quantity_requested = Decimal(data.get('stake_quantity_requested')) if data.get('stake_quantity_requested') != None else None,
            payout_quantity_filled = Decimal(data.get('payout_quantity_filled')) if data.get('payout_quantity_filled') != None else None,
            stake_quantity_filled = Decimal(data.get('stake_quantity_filled')) if data.get('stake_quantity_filled') != None else None,
            payout_quantity_cancelled = Decimal(data.get('payout_quantity_cancelled')) if data.get('payout_quantity_cancelled') != None else None,
            stake_quantity_cancelled = Decimal(data.get('stake_quantity_cancelled')) if data.get('stake_quantity_cancelled') != None else None,
            payout_quantity_lapsed = Decimal(data.get('payout_quantity_lapsed')) if data.get('payout_quantity_lapsed') != None else None,
            stake_quantity_lapsed = Decimal(data.get('stake_quantity_lapsed')) if data.get('stake_quantity_lapsed') != None else None,
            payout_quantity_voided = Decimal(data.get('payout_quantity_voided')) if data.get('payout_quantity_voided') != None else None,
            stake_quantity_voided = Decimal(data.get('stake_quantity_voided')) if data.get('stake_quantity_voided') != None else None,
            payout_quantity_remaining = Decimal(data.get('payout_quantity_remaining')) if data.get('payout_quantity_remaining') != None else None,
            stake_quantity_remaining = Decimal(data.get('stake_quantity_remaining')) if data.get('stake_quantity_remaining') != None else None,
            average_matched_price = Decimal(data.get('average_matched_price')) if data.get('average_matched_price') != None else None,
            removal_reason = data.get('removal_reason'),
            matched_timestamp = datetime.strptime(data.get('matched_timestamp'), "%Y-%m-%dT%H:%M:%S.%fZ") if data.get('matched_timestamp') != None else None,
            lapse_timestamp = datetime.strptime(data.get('lapse_timestamp'), "%Y-%m-%dT%H:%M:%S.%fZ") if data.get('lapse_timestamp') != None else None,
            cancel_timestamp = datetime.strptime(data.get('cancel_timestamp'), "%Y-%m-%dT%H:%M:%S.%fZ") if data.get('cancel_timestamp') != None else None,
            created = datetime.strptime(data.get('created'), "%Y-%m-%dT%H:%M:%S.%fZ"),
            is_removed = data.get('is_removed'),
            is_open = data.get('is_open'),
            is_matched = data.get('is_matched'),
            is_settled = data.get('is_settled'),
        )

    @property
    def selection(self):
        return self.state.selections.get(
            self.selection_id
        )
    
    @property
    def market(self):
        return self.state.markets.get(
            self.market_id
        )

    @property
    def pool(self):
        if self.market:
            return self.market.pool
        return None

    @property
    def pool_id(self):
        if not self._pool_id:
            market = self.state.markets.get(self.market_id)  
            if market:
                self._pool_id = market.pool_id
        return self._pool_id


from .pool_profile import PoolProfile
from decimal import Decimal
from ..position import Position
from ..order import OrderSide
from typing import List


class MatchedPositionProfile(PoolProfile):

    # Inherits all necessary functionality 
    # from the parent PoolProfile class

    @classmethod
    def generate_from_positions(
        cls,
        pool,
        positions: List[Position] = None,
    ):
        instance = cls(
            pool = pool
        )

        if positions == None and pool:
            positions = pool.positions

        for position in positions:
            if position.net_payout > 0:
                side = OrderSide.BUY
                quantity = Decimal(position.net_payout)
            else:
                side = OrderSide.SELL
                quantity = Decimal(-position.net_payout)
            instance.add(
                market_id = int(position.market_id), 
                selection_id = int(position.selection_id),
                side = side,
                price = Decimal(position.new_stake / position.net_payout) if position.net_payout != 0 else Decimal(0),
                quantity = quantity
            )
        return instance

    def print(
        self
    ):
        print('='*40)
        print('POSITION PROFILE')
        print(f'Pool: {self.pool.id}')
        print('-'*40)
        for market_id in self.data.keys():
            print(f'MARKET: {market_id}')
            for selection_id in self.data[market_id].keys():
                print(f' - SELECTION: {selection_id}')
                for side in self.data[market_id][selection_id].keys():
                    print(f'   - {side.value.upper()}')
                    for price, quantity in sorted(self.data[market_id][selection_id][side].items()):
                        print(f'      {price:.3f}: {quantity:8.2f}')
            print('-'*40)
        print('='*40)